/*
 * Copyright (c) 2016 Samsung Electronics Co., Ltd.
 *
 * Licensed under the Apache License, Version 2.0 (the License);
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 * Authors: Adrian Szyndela <adrian.s@samsung.com>
 *          Łukasz Stelmach <l.stelmach@samsung.com>
 */
/**
 * @file crash-stack-aarch64.c
 * @brief unwinding call stacks, functions specific for aarch64
 */
#include "crash-stack.h"
#include <sys/user.h>
#include <string.h>

static struct user_regs_struct g_registers;	///< static storage for ptrace buffer for registers

/**
 * @brief Important registers for unwinding stack on aarch64
 */
struct Regs {
	Dwarf_Addr x29;		///< x29 register contains frame pointer
	Dwarf_Addr x30;		///< x30 register contains return address
	Dwarf_Addr pc;		///< register: program counter
	Dwarf_Addr sp;		///< register: stack pointer
};

#define REG_SP 32		///< number of stack pointer register
#define REG_PC 33		///< number of program counter register
#define REG_X29 29		///< number of x29 register
#define REG_X30 30		///< number of x30 register

typedef struct Regs Regs;	///< convenience type definition
static Regs g_regs;			///< static storage for target register values

void *_crash_stack_get_memory_for_ptrace_registers(size_t *size)
{
	if (NULL != size)
		*size = sizeof(g_registers);
	return &g_registers;
}

void _crash_stack_set_ptrace_registers(void *regbuf)
{
  struct user_regs_struct *regs = regbuf;

  memcpy(_get_place_for_register_value("sp", 0), &regs->sp, sizeof(regs->sp));
  memcpy(_get_place_for_register_value("pc", 0), &regs->pc, sizeof(regs->pc));
  memcpy(_get_place_for_register_value("x29", 0), &regs->regs[29], sizeof(regs->regs[29]));
  memcpy(_get_place_for_register_value("x30", 0), &regs->regs[30], sizeof(regs->regs[30]));
}

void _create_crash_stack(Dwfl *dwfl, Elf *core, pid_t pid, Mappings *mappings, Callstack *callstack)
{
	callstack->elems = 0;
	callstack->proc[callstack->elems++].addr = g_regs.pc;
	callstack->proc[callstack->elems++].addr = g_regs.x30;

	bool end = false;

	do {
		uint64_t newx29, newx30;
		bool read29 = _crash_stack_libelf_read_value(dwfl, core, pid,
				g_regs.x29,
				&newx29, sizeof(newx29), mappings);
		bool read30 = _crash_stack_libelf_read_value(dwfl, core, pid,
				g_regs.x29 + sizeof(newx29),
				&newx30, sizeof(newx30), mappings);
		if (read29 && read30) {
			callstack->proc[callstack->elems++].addr = newx30;
			g_regs.x29 = newx29;
		}
		else end = true;
	} while (!end);
}

void *_get_place_for_register_value(const char *regname, int regnum)
{
	if (strcmp(regname, "pc") == 0 || REG_PC == regnum)
		return &g_regs.pc;
	else if (strcmp(regname, "sp") == 0 || REG_SP == regnum)
		return &g_regs.sp;
	else if (strcmp(regname, "x29") == 0 || REG_X29 == regnum)
		return &g_regs.x29;
	else if (strcmp(regname, "x30") == 0 || REG_X30 == regnum)
		return &g_regs.x30;
	return NULL;
}

#define PSR_MODE32_BIT  0x00000010
#define PSR_MODE_MASK   0x0000000f

void _crash_stack_print_regs(FILE* outputfile)
{
	int i;

	fprintf(outputfile, "\nRegister Information\n");
	for (i = 0; i <= 30; i++) {
		fprintf(outputfile, "x%-2d  = 0x%016llx", i, g_registers.regs[i]);
		if (i != 30 && 0 == i % 2)
			fprintf(outputfile, ", ");
		else
			fprintf(outputfile, "\n");
	}
	fprintf(outputfile, "\n");
	fprintf(outputfile, "xr   = 0x%016llx, ip0  = 0x%016llx\nip1  = 0x%016llx, pr   = 0x%016llx\n",
		g_registers.regs[8],   /* Indirect result location register */
		g_registers.regs[16],  /* Intra-procedure call scratch register 0 */
		g_registers.regs[17],  /* Intra-procedure call scratch register 1 */
		g_registers.regs[18]); /* Platform register */
	fprintf(outputfile, "fp   = 0x%016llx, lr   = 0x%016llx\npc   = 0x%016llx, sp   = 0x%016llx\n",
		g_registers.regs[29],  /* Frame register */
		g_registers.regs[30],  /* Procedure link register */
		g_registers.pc,        /* Program counter */
		g_registers.sp);       /* Stack pointer */
}
